import asyncio
import logging
import os
import webbrowser
from threading import Lock

from aiohttp import ClientSession
from aiohttp.client_exceptions import (
    ClientConnectorError,
    ClientOSError,
    ServerDisconnectedError,
)

from .errors import NoServerFoundError
from .forking import forkserver
from .loops import get_loop
from .utils import current_origin, url

log = logging.getLogger(__name__)
fork_lock = Lock()

websockets = {}
sessions = {}

websocket_options = {"max_msg_size": 1024 * 1024 * 1024}


async def websocket():
    ws, _ = await websocket_state()
    return ws


async def websocket_state():
    # Here we get the magic cookie of our current thread in current pid
    origin = current_origin()
    if origin in websockets:
        return websockets[origin], True

    sessions[origin] = sessions.get(origin, ClientSession())

    with fork_lock:
        try:
            ws = await sessions[origin].ws_connect(
                url("back"), **websocket_options
            )
            log.info("Found existing kalong server")
        except ClientConnectorError:
            # If there are no server available, fork one
            log.info("No kalong server, starting one")
            forkserver()
            for _ in range(500):
                await asyncio.sleep(0.05)
                try:
                    ws = await sessions[origin].ws_connect(
                        url("back"), **websocket_options
                    )
                    break
                except ClientConnectorError:
                    pass
            else:
                raise NoServerFoundError()

        # webbrowser.open should be in the mutex too, it's not thread safe
        if os.getenv("KALONG_NO_BROWSER") or not webbrowser.open(url("front")):
            log.warn(
                "Please open your browser to the following url: "
                f'{url("front")}'
            )

    websockets[origin] = ws
    return ws, False


async def close_websocket():
    origin = current_origin()
    try:
        if origin in websockets:
            await websockets[origin].close()
    finally:
        if origin in websockets:
            del websockets[origin]
        if origin in sessions:
            try:
                await sessions[origin].close()
            finally:
                del sessions[origin]


def die():
    log.info("Dying, closing socket.")
    loop = get_loop()
    if loop.is_running():
        loop.create_task(close_websocket())
    else:
        loop.run_until_complete(close_websocket())


async def close_all(closeables):
    if not closeables:
        return
    await asyncio.gather(
        *[closeable.close() for closeable in closeables.values()],
        return_exceptions=True,
    )


def clean_websockets():
    log.info(
        f"Cleaning at exit {len(websockets)} ws and {len(sessions)} sessions"
    )
    asyncio.run(close_all(websockets))
    asyncio.run(close_all(sessions))



import logging
import sys
import threading
from multiprocessing import process
from pathlib import Path

from .communication import communicate
from .stepping import steppings, stop_trace
from .websockets import die

log = logging.getLogger(__name__)

kalong_dir = str(Path(__file__).resolve().parent)


def trace(origin, frame, event, arg):
    stepping = steppings.get(origin)
    if not stepping:
        return

    type = stepping.get("type")
    originalFrame = stepping.get("frame")
    lno = stepping.get("lno")
    filename = frame.f_code.co_filename

    # When we are in _shutdown of thread, program is finished
    if (
        filename == threading.__file__
        and frame.f_code.co_name in ["_shutdown", "_bootstrap_inner"]
        or filename == process.__file__
        and frame.f_code.co_name == "_bootstrap"
    ):
        stop_trace(frame)
        die()
        return

    # If we are tracing, we continue tracing if this is not an exception
    if type == "trace" and event != "exception":
        return sys.gettrace()

    # If we are continuing, don't trace anything but current frame
    if type == "continue":
        if frame == originalFrame:
            # Continue tracing current frame
            if event != "exception":
                return sys.gettrace()
        else:
            # Stop tracing if below
            return

    # Don't trace importlib bootstrapping
    if filename == "<frozen importlib._bootstrap>":
        return

    # Don't trace self
    if filename.startswith(kalong_dir):
        return

    # Don't trace under frames if we are not stepping 'into'
    if event == "call" and type != "stepInto":
        return

    if (
        # Trace: This is an exception: stop
        type == "trace"
        # Continue: This is an exception at current frame: stop
        or type == "continue"
        # Step: Normal stepping
        or type == "step"
        # StepInto: Stepping normally in new frame as well
        or type == "stepInto"
        # StepUntil: Stepping only if line number is greater than previously
        or (
            type == "stepUntil"
            and (frame != originalFrame or frame.f_lineno > lno)
        )
        # StepOut: Stepping only if we are returning or in the parent frame
        or (
            type == "stepOut"
            and (
                (event == "return" and frame == originalFrame)
                or event == "line"
                and frame == originalFrame.f_back
            )
        )
    ):
        # Enter the websocket communication loop that pauses the execution
        communicate(frame, event, arg)

    return sys.gettrace()



"""A new take on debugging"""
__version__ = "0.2.3"
import os
import sys

from .config import Config

config = Config()
config.from_env()


def breakpoint():
    from .stepping import add_step, start_trace, stop_trace

    frame = sys._getframe().f_back
    stop_trace(frame)
    add_step("step", frame)
    start_trace(frame)


def break_above(level):
    from .stepping import add_step, start_trace, stop_trace

    frame = sys._getframe()
    while level > 0 and frame.f_back:
        level += 1
        frame = frame.f_back

    stop_trace(frame)
    add_step("step", frame)
    start_trace(frame)


def start_trace(break_=False, full=False):
    from .stepping import add_step, start_trace

    frame = sys._getframe().f_back
    add_step(
        "stepInto" if break_ else ("trace" if full else "continue"), frame
    )
    start_trace(frame)


def stop_trace():
    from .stepping import stop_trace

    frame = sys._getframe().f_back
    stop_trace(frame)


class trace:
    def __init__(self, break_=False, full=False):
        self.break_ = break_
        self.full = full

    def __enter__(self):
        start_trace(self.break_, self.full)

    def __exit__(self, *args):
        stop_trace()


def run_file(filename, *args):
    from .utils import fake_argv

    # Cleaning __main__ namespace
    import __main__

    __main__.__dict__.clear()
    __main__.__dict__.update(
        {
            "__name__": "__main__",
            "__file__": filename,
            "__builtins__": __builtins__,
        }
    )
    with open(filename, "rb") as fp:
        statement = compile(
            fp.read(), os.path.abspath(os.path.normcase(filename)), "exec"
        )

    globals = __main__.__dict__
    locals = globals
    with fake_argv(filename, *args):
        with trace(break_=True):
            exec(statement, globals, locals)


def shell():
    # Launch a shell
    from .communication import communicate

    frame = sys._getframe()
    # Enter the websocket communication loop that pauses the execution
    communicate(frame, "shell", [])



import asyncio
import json
import linecache
import logging
import threading

from aiohttp import WSMsgType

from . import config
from .debugger import (
    get_frame,
    get_title,
    serialize_answer,
    serialize_diff_eval,
    serialize_frames,
    serialize_inspect,
    serialize_inspect_eval,
    serialize_suggestion,
)
from .loops import get_loop
from .stepping import add_step, clear_step, stop_trace
from .utils import basicConfig, get_file_from_code
from .websockets import die, websocket_state
from .errors import SetFrameError

log = logging.getLogger(__name__)
basicConfig(level=config.log_level)


def communicate(frame, event, arg):
    loop = get_loop()
    if loop.is_running():
        raise SetFrameError(frame, event, arg)

    try:
        loop.run_until_complete(communication_loop(frame, event, arg))
    except asyncio.CancelledError:
        log.info("Loop got cancelled")
        die()


async def init(ws, frame, event, arg):
    tb = arg[2] if event == "exception" else None

    await ws.send_json({"type": "SET_THEME", "theme": event})
    await ws.send_json(
        {"type": "SET_TITLE", "title": get_title(frame, event, arg)}
    )
    await ws.send_json(
        {
            "type": "SET_FRAMES",
            "frames": list(serialize_frames(frame, tb))
            if event != "shell"
            else [],
        }
    )


async def communication_loop(frame_, event_, arg_):
    frame = frame_
    event = event_
    arg = arg_

    ws, existing = await websocket_state()
    await ws.send_json({"type": "PAUSE"})

    if existing:
        # If the socket is already opened we need to update client state
        await init(ws, frame, event, arg)
        # Otherwise if it's new, just wait for HELLO to answer current state

    stop = False
    async for msg in ws:
        if msg.type == WSMsgType.TEXT:
            data = json.loads(msg.data)
            if data["type"] == "HELLO":
                await init(ws, frame, event, arg)
                response = {
                    "type": "SET_INFO",
                    "config": config.__dict__,
                    "main": threading.current_thread()
                    is threading.main_thread(),
                }

            elif data["type"] == "GET_FILE":
                filename = data["filename"]
                file = "".join(linecache.getlines(filename))
                if not file:
                    file = get_file_from_code(frame, filename)
                response = {
                    "type": "SET_FILE",
                    "filename": filename,
                    "source": file,
                }

            elif data["type"] == "SET_PROMPT":
                try:
                    response = {
                        "type": "SET_ANSWER",
                        "key": data["key"],
                        "command": data.get("command"),
                        "frame": data.get("frame"),
                        **serialize_answer(
                            data["prompt"], get_frame(frame, data.get("frame"))
                        ),
                    }
                except SetFrameError as e:
                    frame = e.frame
                    event = e.event
                    arg = e.arg

                    await init(ws, frame, event, arg)

                    response = {
                        "type": "SET_ANSWER",
                        "key": data["key"],
                        "command": data.get("command"),
                        "frame": data.get("frame"),
                        "prompt": data["prompt"].strip(),
                        "answer": "",
                        "duration": 0,
                    }

            elif data["type"] == "REQUEST_INSPECT":
                response = {
                    "type": "SET_ANSWER",
                    "key": data["key"],
                    "command": data.get("command"),
                    **serialize_inspect(data["id"]),
                }

            elif data["type"] == "REQUEST_INSPECT_EVAL":
                response = {
                    "type": "SET_ANSWER",
                    "key": data["key"],
                    "command": data.get("command"),
                    **serialize_inspect_eval(
                        data["prompt"], get_frame(frame, data.get("frame"))
                    ),
                }

            elif data["type"] == "REQUEST_DIFF_EVAL":
                response = {
                    "type": "SET_ANSWER",
                    "key": data["key"],
                    "command": data.get("command"),
                    **serialize_diff_eval(
                        data["left"],
                        data["right"],
                        get_frame(frame, data.get("frame")),
                    ),
                }

            elif data["type"] == "REQUEST_SUGGESTION":
                response = {
                    "type": "SET_SUGGESTION",
                    **serialize_suggestion(
                        data["prompt"],
                        data["from"],
                        data["to"],
                        data["cursor"],
                        get_frame(frame, data.get("frame")),
                    ),
                }

            elif data["type"] == "DO_COMMAND":
                command = data["command"]
                response = {"type": "ACK", "command": command}
                if command == "run":
                    clear_step()
                    stop_trace(frame)
                elif command == "stop":
                    clear_step()
                    stop_trace(frame)
                    die()
                else:
                    add_step(command, frame)
                stop = True

            else:
                response = {
                    "type": "error",
                    "message": f"Unknown type {data['type']}",
                }

            log.debug(f"Got {data} answering with {response}")
            response["local"] = True
            await ws.send_json(response)

            if stop:
                break

        elif msg.type == WSMsgType.ERROR:
            log.error(f"WebSocket closed", exc_info=ws.exception())
            break

    # Browser exited, stopping debug if we are not stepping
    if not stop:
        stop_trace(frame)



import os
from pathlib import Path
from subprocess import run

from . import config, run_file, shell

os.environ["PYTHONBREAKPOINT"] = "kalong.breakpoint"
config.from_args()

if config.server:
    from .server import serve

    serve()

elif config.inject:
    kalong_dir = Path(__file__).resolve().parent.parent
    gdb_command = ["gdb", "-p", str(config.inject), "-batch"] + [
        "-eval-command=call %s" % hook
        for hook in [
            "(int) PyGILState_Ensure()",  # Getting the GIL
            '(int) PyRun_SimpleString("'
            f"print('* Kalong injection from {os.getpid()} *');"
            "import sys;"  # Putting kalong project directory in sys path:
            f"sys.path.insert(0, '{kalong_dir}');"
            "import kalong;"  # Setting breakpoint:
            "kalong.break_above(2);"
            '")',
            # Releasing the GIL with the PyGILState_Ensure handle:
            "(void) PyGILState_Release($1)",
        ]
    ]
    print(f'Running: {" ".join(gdb_command)}')
    run(gdb_command)

else:
    if config.command:
        run_file(*config.command)
    else:
        shell()



import asyncio
import json
import logging
import os
import signal
from itertools import chain
from pathlib import Path

from aiohttp import WSMsgType, web
from aiohttp.web_runner import GracefulExit

from . import config
from .errors import NoClientFoundError
from .utils import basicConfig, USER_SIGNAL, human_readable_side, parse_origin
from .websockets import websocket_options

log = logging.getLogger(__name__)
basicConfig(level=config.log_level)


def maybe_bail(app):
    if not app["front"] and not app["back"]:
        log.info("No remaining clients, gracefully bailing.")
        raise GracefulExit()
    else:
        log.info(
            f"Not closing since {len(app['front'])} front and "
            + f"{len(app['back'])} back connections remains."
        )


async def shutdown(app):
    log.info("App shutdown, closing remaining websockets.")
    await asyncio.gather(
        *[
            ws.close()
            for ws in chain(app["front"].values(), app["back"].values())
        ]
    )
    app["front"].clear()
    app["back"].clear()


async def peer(app, side, origin):
    if origin in app[side]:
        return app[side][origin]
    else:
        for _ in range(500):
            await asyncio.sleep(0.05)
            if origin in app[side]:
                return app[side][origin]
        raise NoClientFoundError(
            f"No {human_readable_side(side)} was open to connect to debugger"
        )


async def websocket(request):
    side = request.match_info["side"]
    other_side = "back" if side == "front" else "front"
    origin = request.match_info["origin"]
    ws = web.WebSocketResponse(**websocket_options)

    state = ws.can_prepare(request)
    if not state.ok:
        log.debug(f"Sending {side} app for {origin}")
        return web.FileResponse(
            Path(__file__).parent / "static" / "index.html"
        )

    await ws.prepare(request)

    log.debug(f"{side.title()} app connected for {origin}")
    request.app[side][origin] = ws

    async for msg in ws:
        if msg.type == WSMsgType.TEXT:
            data = json.loads(msg.data)
            log.debug(f"{side} -> {other_side}: {data}")
            try:
                pair = await peer(request.app, other_side, origin)
            except NoClientFoundError:
                log.error(
                    f"No {human_readable_side(other_side)} connection was found, aborting debugger"
                )
                break
            if data["type"] == "DO_COMMAND":
                if data["command"] in ["pause", "kill"]:
                    _, pid, _ = parse_origin(origin)
                    os.kill(
                        pid,
                        signal.SIGTERM
                        if data["command"] == "kill"
                        else USER_SIGNAL,
                    )
                    continue

            await pair.send_json(data)
        elif msg.type == WSMsgType.ERROR:
            log.error(f"{side.title()} closed", exc_info=ws.exception())

    log.debug(f"Closing {side}")
    await ws.close()
    del request.app[side][origin]

    if not config.detached:
        if origin in request.app[other_side]:
            log.debug(f"Closing {other_side} due to {side} closing.")
            await request.app[other_side][origin].close()

        if origin in request.app[other_side]:
            del request.app[other_side][origin]

    maybe_bail(request.app)
    return ws


def serve():
    app = web.Application()
    app["front"] = {}
    app["back"] = {}
    app.on_shutdown.append(shutdown)
    app.router.add_get(r"/{side:(front|back)}/{origin}", websocket)
    app.router.add_static(
        "/assets/", Path(__file__).parent / "static" / "assets"
    )
    web.run_app(app, host=config.host, port=config.port, print=False)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kalong', 'kalong.utils', 'kalong.utils.doc_lookup']

package_data = \
{'': ['*'], 'kalong': ['static/*', 'static/assets/*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0', 'jedi>=0.18.1,<0.19.0']

setup_kwargs = {
    'name': 'kalong',
    'version': '0.2.3',
    'description': 'A new take on python debugging',
    'long_description': None,
    'author': 'Florian Mounier',
    'author_email': 'paradoxxx.zero@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/paradoxxxzero/kalong',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import linecache
import os
import re
import signal
import socket
import sys
import threading
from pathlib import Path

from .. import config
from .iterators import iter_frame

try:
    from log_colorizer import basicConfig
except ImportError:
    from logging import basicConfig  # noqa

try:
    import uncompyle6
except ImportError:
    uncompyle6 = None

ORIGIN_RE = re.compile(r"(.+)__(.+)--(.+)")


def current_origin():
    return f"{socket.getfqdn()}__{os.getpid()}--{threading.get_ident()}"


def parse_origin(origin):
    match = ORIGIN_RE.search(origin)
    if not match:
        raise ValueError(
            f"Invalid origin format {origin} must match {ORIGIN_RE}"
        )
    fqdn, pid, ident = match.groups()

    return fqdn, int(pid), int(ident)


class fake_argv:
    def __init__(self, *args):
        self.args = args

    def __enter__(self):
        self._old_args = sys.argv
        sys.argv = self.args

    def __exit__(self, *err):
        sys.argv = self._old_args


USER_SIGNAL = signal.SIGILL if sys.platform == "win32" else signal.SIGUSR1


def url(side):
    origin = current_origin()
    overriden_port = config.front_port if side == "front" else config.port
    protocol = config.protocol
    host = config.host
    return f"{protocol}://{host}:{overriden_port}/{side}/{origin}"


def human_readable_side(side):
    return "browser" if side == "front" else "program"


def discompile(code):
    if not uncompyle6:
        return ""
    try:
        return uncompyle6.deparse_code2str(code)
    except Exception:
        return ""


def get_file_from_code(frame, filename):
    for frame in iter_frame(frame):
        co_filename = frame.f_code.co_filename or "<unspecified>"
        if co_filename == "<frozen importlib._bootstrap>":
            co_filename = os.path.join(
                os.path.dirname(linecache.__file__),
                "importlib",
                "_bootstrap.py",
            )
        fn = Path(co_filename)
        if not co_filename.startswith("<"):
            fn = fn.resolve()

        if filename == str(fn):
            return discompile(frame.f_code)
